package be.iminds.ilabt.jfed.util;

import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.net.SocketFactory;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.*;
import java.nio.channels.SocketChannel;
import java.util.List;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.atomic.AtomicBoolean;

public class LoggingSocketFactory extends SocketFactory {
    private static final Logger LOG = LoggerFactory.getLogger(LoggingSocketFactory.class);

    @Nonnull private final SocketFactory realSocketFactory;
    @Nonnull private final List<SocketLogger> loggers;

    public interface SocketLogger {
        void onConnectStart();
        void onConnectEnd();
        void onFirstReadStart();
        void onFirstReadEnd();
        void onFirstWriteStart();
        void onFirstWriteEnd();
        void onClose();
    }

    public LoggingSocketFactory(@Nonnull SocketFactory realSocketFactory) {
        this.realSocketFactory = realSocketFactory;
        loggers = new CopyOnWriteArrayList<>();
    }

    public void addSocketLogger(@Nonnull SocketLogger socketLogger) {
        loggers.add(socketLogger);
    }
    public void removeSocketLogger(@Nonnull SocketLogger socketLogger) {
        loggers.remove(socketLogger);
    }

    protected interface SocketLoggerFire {
        void fire(SocketLogger socketLogger);
    }
    protected void fire(SocketLoggerFire f) {
        loggers.stream().forEach((l) -> {
            try {
                f.fire(l);
            } catch (Exception e) {
                LOG.error("Exception logging socket action.", e);
            }
        });
    }

    @Override
    public Socket createSocket() throws IOException {
        return new LoggingSocket(realSocketFactory.createSocket());
    }

    @Override
    public Socket createSocket(String s, int i) throws IOException, UnknownHostException {
        return new LoggingSocket(realSocketFactory.createSocket(s, i));
    }

    @Override
    public Socket createSocket(String s, int i, InetAddress inetAddress, int i1) throws IOException, UnknownHostException {
        return new LoggingSocket(realSocketFactory.createSocket(s, i, inetAddress, i1));
    }

    @Override
    public Socket createSocket(InetAddress inetAddress, int i) throws IOException {
        return new LoggingSocket(realSocketFactory.createSocket(inetAddress, i));
    }

    @Override
    public Socket createSocket(InetAddress inetAddress, int i, InetAddress inetAddress1, int i1) throws IOException {
        return new LoggingSocket(realSocketFactory.createSocket(inetAddress, i, inetAddress1, i1));
    }

    private class LoggingSocket extends Socket {
        private final Socket realSocket;

        private AtomicBoolean firstRead = new AtomicBoolean(false);
        private AtomicBoolean firstWrite = new AtomicBoolean(false);

        public LoggingSocket(Socket realSocket) {
            this.realSocket = realSocket;
        }

        @Override
        public void connect(SocketAddress socketAddress) throws IOException {
            fire(SocketLogger::onConnectStart);
            realSocket.connect(socketAddress);
            fire(SocketLogger::onConnectEnd);
        }

        @Override
        public void connect(SocketAddress socketAddress, int i) throws IOException {
            fire(SocketLogger::onConnectStart);
            realSocket.connect(socketAddress, i);
            fire(SocketLogger::onConnectEnd);
        }

        @Override
        public void bind(SocketAddress socketAddress) throws IOException {
            realSocket.bind(socketAddress);
        }

        @Override
        public InetAddress getInetAddress() {
            return realSocket.getInetAddress();
        }

        @Override
        public InetAddress getLocalAddress() {
            return realSocket.getLocalAddress();
        }

        @Override
        public int getPort() {
            return realSocket.getPort();
        }

        @Override
        public int getLocalPort() {
            return realSocket.getLocalPort();
        }

        @Override
        public SocketAddress getRemoteSocketAddress() {
            return realSocket.getRemoteSocketAddress();
        }

        @Override
        public SocketAddress getLocalSocketAddress() {
            return realSocket.getLocalSocketAddress();
        }

        @Override
        public SocketChannel getChannel() {
            return realSocket.getChannel();
        }

        @Override
        public InputStream getInputStream() throws IOException {
            return realSocket.getInputStream();
        }

        @Override
        public OutputStream getOutputStream() throws IOException {
            return realSocket.getOutputStream();
        }

        @Override
        public void setTcpNoDelay(boolean b) throws SocketException {
            realSocket.setTcpNoDelay(b);
        }

        @Override
        public boolean getTcpNoDelay() throws SocketException {
            return realSocket.getTcpNoDelay();
        }

        @Override
        public void setSoLinger(boolean b, int i) throws SocketException {
            realSocket.setSoLinger(b, i);
        }

        @Override
        public int getSoLinger() throws SocketException {
            return realSocket.getSoLinger();
        }

        @Override
        public void sendUrgentData(int i) throws IOException {
            realSocket.sendUrgentData(i);
        }

        @Override
        public void setOOBInline(boolean b) throws SocketException {
            realSocket.setOOBInline(b);
        }

        @Override
        public boolean getOOBInline() throws SocketException {
            return realSocket.getOOBInline();
        }

        @Override
        public synchronized void setSoTimeout(int i) throws SocketException {
            realSocket.setSoTimeout(i);
        }

        @Override
        public synchronized int getSoTimeout() throws SocketException {
            return realSocket.getSoTimeout();
        }

        @Override
        public synchronized void setSendBufferSize(int i) throws SocketException {
            realSocket.setSendBufferSize(i);
        }

        @Override
        public synchronized int getSendBufferSize() throws SocketException {
            return realSocket.getSendBufferSize();
        }

        @Override
        public synchronized void setReceiveBufferSize(int i) throws SocketException {
            realSocket.setReceiveBufferSize(i);
        }

        @Override
        public synchronized int getReceiveBufferSize() throws SocketException {
            return realSocket.getReceiveBufferSize();
        }

        @Override
        public void setKeepAlive(boolean b) throws SocketException {
            realSocket.setKeepAlive(b);
        }

        @Override
        public boolean getKeepAlive() throws SocketException {
            return realSocket.getKeepAlive();
        }

        @Override
        public void setTrafficClass(int i) throws SocketException {
            realSocket.setTrafficClass(i);
        }

        @Override
        public int getTrafficClass() throws SocketException {
            return realSocket.getTrafficClass();
        }

        @Override
        public void setReuseAddress(boolean b) throws SocketException {
            realSocket.setReuseAddress(b);
        }

        @Override
        public boolean getReuseAddress() throws SocketException {
            return realSocket.getReuseAddress();
        }

        @Override
        public synchronized void close() throws IOException {
            realSocket.close();
            fire(SocketLogger::onClose);
        }

        @Override
        public void shutdownInput() throws IOException {
            realSocket.shutdownInput();
        }

        @Override
        public void shutdownOutput() throws IOException {
            realSocket.shutdownOutput();
        }

        @Override
        public boolean isConnected() {
            return realSocket.isConnected();
        }

        @Override
        public boolean isBound() {
            return realSocket.isBound();
        }

        @Override
        public boolean isClosed() {
            return realSocket.isClosed();
        }

        @Override
        public boolean isInputShutdown() {
            return realSocket.isInputShutdown();
        }

        @Override
        public boolean isOutputShutdown() {
            return realSocket.isOutputShutdown();
        }

        @Override
        public void setPerformancePreferences(int i, int i1, int i2) {
            realSocket.setPerformancePreferences(i, i1, i2);
        }


        private class LoggerInputStream extends InputStream {
            private InputStream realInputStream;

            public LoggerInputStream(InputStream realInputStream) {
                this.realInputStream = realInputStream;
            }

            @Override
            public int read(@NotNull byte[] bytes) throws IOException {
                boolean isFirstRead = firstRead.compareAndSet(false, true);
                if (isFirstRead)
                    fire(SocketLogger::onFirstReadStart);
                int res = realInputStream.read(bytes);
                if (isFirstRead)
                    fire(SocketLogger::onFirstReadEnd);
                return res;
            }

            @Override
            public int read(@NotNull byte[] bytes, int i, int i1) throws IOException {
                boolean isFirstRead = firstRead.compareAndSet(false, true);
                if (isFirstRead)
                    fire(SocketLogger::onFirstReadStart);
                int res = realInputStream.read(bytes, i, i1);
                if (isFirstRead)
                    fire(SocketLogger::onFirstReadEnd);
                return res;
            }

            @Override
            public long skip(long l) throws IOException {
                return realInputStream.skip(l);
            }

            @Override
            public int available() throws IOException {
                return realInputStream.available();
            }

            @Override
            public void close() throws IOException {
                realInputStream.close();
            }

            @Override
            public synchronized void mark(int i) {
                realInputStream.mark(i);
            }

            @Override
            public synchronized void reset() throws IOException {
                realInputStream.reset();
            }

            @Override
            public boolean markSupported() {
                return realInputStream.markSupported();
            }

            @Override
            public int read() throws IOException {
                boolean isFirstRead = firstRead.compareAndSet(false, true);
                if (isFirstRead)
                    fire(SocketLogger::onFirstReadStart);
                int res = realInputStream.read();
                if (isFirstRead)
                    fire(SocketLogger::onFirstReadEnd);
                return res;
            }
        }
        
        private class LoggerOutputStream extends OutputStream {
            private OutputStream realOutputStream;

            public LoggerOutputStream(OutputStream realOutputStream) {
                this.realOutputStream = realOutputStream;
            }

            @Override
            public void write(@NotNull byte[] bytes) throws IOException {
                boolean isFirstWrite = firstWrite.compareAndSet(false, true);
                if (isFirstWrite)
                    fire(SocketLogger::onFirstWriteStart);
                realOutputStream.write(bytes);
                if (isFirstWrite)
                    fire(SocketLogger::onFirstWriteEnd);
            }

            @Override
            public void write(@NotNull byte[] bytes, int i, int i1) throws IOException {
                boolean isFirstWrite = firstWrite.compareAndSet(false, true);
                if (isFirstWrite)
                    fire(SocketLogger::onFirstWriteStart);
                realOutputStream.write(bytes, i, i1);
                if (isFirstWrite)
                    fire(SocketLogger::onFirstWriteEnd);
            }

            @Override
            public void write(int i) throws IOException {
                boolean isFirstWrite = firstWrite.compareAndSet(false, true);
                if (isFirstWrite)
                    fire(SocketLogger::onFirstWriteStart);
                realOutputStream.write(i);
                if (isFirstWrite)
                    fire(SocketLogger::onFirstWriteEnd);
            }

            @Override
            public void flush() throws IOException {
                realOutputStream.flush();
            }

            @Override
            public void close() throws IOException {
                realOutputStream.close();
            }
        }
    }
}
